from __future__ import annotations
import inspect
import logging
from asyncio import sleep
from sys import exc_info
from functools import wraps
from discord import Forbidden
from red_star.rs_errors import ChannelNotFoundError, CommandSyntaxError, UserPermissionError
from red_star.rs_utils import respond, sub_user_data

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import discord
    import plugin_manager
    from red_star.client import RedStar
    from channel_manager import ChannelManager
    from builtins import function


class CommandDispatcher:
    def __init__(self, client: RedStar, guild: discord.Guild, channel_manager: ChannelManager):
        self.client = client
        self.guild = guild
        self.channel_manager = channel_manager
        self.config_manager = client.config_manager
        self.logger = logging.getLogger(f"red_star.command_dispatcher.{guild.id}")
        self.config = self.config_manager.get_server_config(self.guild, "command_dispatcher",
                                                            default_config={"command_prefix": "!",
                                                                            "permission_overrides": {}})

        self.commands = {}
        self.last_error = None

    async def on_message(self, msg: discord.Message):
        await self.command_check(msg)

    def initialize_command_permissions(self, command: Command):
        overrides = self.config["permission_overrides"]

        if command.name in overrides:
            if "permissions_all" in overrides:
                command.perms.permissions_all = set(overrides["permissions_all"])
            if "optional_permissions" in overrides:
                command.perms.optional_permissions = {k: set(v) for k, v in overrides["optional_permissions"].items()}
            command.perms.permissions_any = set(overrides.get("permissions_any", []))
            command.perms.user_overrides = set(overrides.get("user_overrides", []))
            command.perms.role_overrides = set(overrides.get("role_overrides", []))

        command.perms.bot_maintainers = self.config_manager.config["global"].get("bot_maintainers", [])

    def register_plugin(self, plugin: plugin_manager.BasePlugin):
        for _, mth in inspect.getmembers(plugin, predicate=inspect.ismethod):
            if hasattr(mth, "_command"):
                self.register(mth, mth.name.lower())

    def deregister_plugin(self, plugin: plugin_manager.BasePlugin):
        for _, mth in inspect.getmembers(plugin, predicate=inspect.ismethod):
            if hasattr(mth, "_command"):
                self.deregister(mth, mth.name.lower())

    def register(self, command_func: function, name: str, is_alias=False):
        """
        Register commands in the command list and handle conflicts. If it's an
        alias, we don't want it to overwrite non-alias commands. Otherwise,
        overwrite based on priority, and failing that, load order.

        :param command_func: The command function, with its added information.
        :param name: The command's name, for indexing.
        :param is_alias: A boolean that tells the registrar not to overwrite other commands.
        """
        self.logger.debug(f"Registering command {name} from {command_func.__self__.name}.")

        if not is_alias:
            self.initialize_command_permissions(command_func)

        if name in self.commands:
            existing_func = self.commands[name]
            if command_func.priority >= existing_func.priority and not is_alias:
                self.commands[name] = command_func
                self.logger.warning(f"Command {name} from {command_func.__self__.name} overwrites command"
                                    f" {existing_func.name} from {existing_func.__self__.name}!")
            else:
                self.logger.warning(f"Command {existing_func.name} from {existing_func.__self__.name} overwrites"
                                    f" command {name} from {command_func.__self__.name}!")
        else:
            self.commands[name] = command_func

        if hasattr(command_func, "aliases") and not is_alias:
            for alias in command_func.aliases:
                self.register(command_func, alias.lower(), is_alias=True)

    def deregister(self, command_func: function, name: str, is_alias=False):
        """
        Deregister commands from the command list when their plugin is deactivated.

        :param command_func: The command function, with its added information.
        :param name: The command's name, for indexing.
        :param is_alias: A boolean that tells the registrar not to overwrite other commands.
        """
        self.logger.debug(f"Deregistering command {name} from {command_func.__self__.name}.")

        if name in self.commands:
            existing_func = self.commands[name]
            # Make sure the command isn't another plugin's
            if command_func == existing_func:
                del self.commands[name]
        else:
            self.logger.debug(f"Could not deregister command {name}, no such command!")

        if hasattr(command_func, "aliases") and not is_alias:
            for alias in command_func.aliases:
                self.deregister(command_func, alias.lower(), is_alias=True)

    # noinspection PyBroadException
    async def run_command(self, command_name: str, msg: discord.Message, dm_cmd=False):
        try:
            command_func = self.commands[command_name]
        except KeyError:
            return
        if dm_cmd:
            return  # TODO: DM commands rework.

        else:
            try:
                if not command_func.run_anywhere:
                    try:
                        cmd_channel = self.channel_manager.get_channel("commands")
                        if msg.channel != cmd_channel:
                            return
                    except ChannelNotFoundError:
                        pass
                await command_func(msg)
                if command_func.delete_call:
                    await sleep(1)
                    try:
                        await msg.delete()
                    except Forbidden:
                        pass
            except CommandSyntaxError as e:
                err = e if e else "Invalid syntax."
                if command_func.syntax:
                    deco = self.config["command_prefix"].lower()
                    await respond(msg, f"**WARNING: {err} ANALYSIS: Proper usage: "
                                       f"{deco}{command_func.name} {command_func.syntax}.**")
                else:
                    await respond(msg, f"**WARNING: {err}**")
            except UserPermissionError as e:
                err = f"\nANALYSIS: {e}" if str(e) else ""
                await respond(msg, sub_user_data(msg.author,
                                                 f"**NEGATIVE. INSUFFICIENT PERMISSION: <usernick>. {err}**"))
            except Forbidden:
                await respond(msg, "**NEGATIVE. This unit does not have permission to perform that action.**")
            except ChannelNotFoundError as e:
                await respond(msg, f"**NEGATIVE. Channel type `{e}` is not set on this server.**")
            except Exception:
                self.last_error = exc_info()
                self.logger.exception("Exception occurred in command. ", exc_info=True)
                await respond(msg, "**WARNING: Error occurred while running command.**")

    # Event hooks

    async def command_check(self, msg: discord.Message):
        try:
            deco = self.config["command_prefix"]
            dm_cmd = False
        except AttributeError:  # Oops, it's a DM isn't it
            deco = "!"
            dm_cmd = True
        if msg.author != self.client.user:
            cnt = msg.content
            if cnt.startswith(deco):
                cmd = cnt[len(deco):].split()[0].lower()
                if cmd in self.commands:
                    await self.run_command(cmd, msg, dm_cmd=dm_cmd)


class Command:
    """
    Defines a decorator that encapsulates a chat command. Provides a common
    interface for all commands, including roles, documentation, usage syntax,
    and aliases.

    :param name: The string name of the command, used to call it on Discord. If two commands have the same name,
     the one with a higher priority will override the other. If both have the same priority, the first registered
     overrides the latter.
    :param aliases: Additional, alternate names for the command that can also be used to call it. If another command
     exists with the same name as an alias, the alias is overridden.
    :param perms: A single permission or set of permissions, provided as string names, that the command will require by
     default to be used. Check discord.py documentation for valid permission names.
    :param optional_perms: A dict of optional permissions - the key is the optional permission group name,
     the value is a set of string permission names required to qualify for the group.
    :param doc: Documentation text for the usage of the command on Discord. This will be displayed when the Help
     command is invoked.
    :param syntax: A string demonstrating the syntax of command usage. Do not include the command name.
    :param priority: The command's registration priority. When two commands have the same name, the one with higher
     priority overrides the lower.
    :param delete_call: If True, the bot will automatically delete the message that invoked the command.
    :param run_anywhere: If True, this command ignores the typical requirement that commands can only be invoked in
     a set channel, if that channel is set.
    :param bot_maintainers_only: If True, only users on the bot maintainer list can use the command. Typically used
     for commands that can alter bot function on more than one server.
    :param dm_command: If True, this command can be used in direct messages.
    :param category: The category name that this command will be filed under in the Help command output.
    """
    def __init__(self, name: str, *aliases: str, perms: str | set[str] = None,
                 optional_perms: dict[str, set[str]] = None, doc: str = None, syntax: str = None, priority: int = 0,
                 delete_call: bool = False, run_anywhere: bool = False, bot_maintainers_only: bool = False,
                 dm_command: bool = False, category: str = "other"):
        if syntax is None:
            syntax = ()
        if isinstance(syntax, str):
            syntax = (syntax,)
        if doc is None:
            doc = ""

        if not perms:
            perms = set()
        if isinstance(perms, str):
            perms = {perms}
        self.perms = CommandPermissions(perms, bot_maintainers_only=bot_maintainers_only,
                                        optional_permissions=optional_perms)

        self.name = name
        self.syntax = syntax
        self.human_syntax = " ".join(syntax)
        self.doc = doc
        self.aliases = aliases
        self.priority = priority
        self.delete_call = delete_call
        self.run_anywhere = run_anywhere
        self.category = category
        self.dm_command = dm_command

    def __call__(self, f):
        """
        Whenever a command is called, its handling gets done here.

        :param f: The function the Command decorator is wrapping.
        :return: The now-wrapped command, with all the trappings.
        """
        @wraps(f)
        async def wrapped(s: plugin_manager.BasePlugin, msg: discord.Message):
            if msg.guild is None and self.dm_command:  # The permission check was handled pre-call.
                return await f(s, msg)
            # user_perms = {x for x, y in msg.channel.permissions_for(msg.author) if y}
            # if msg.guild.voice_client:
            #     user_perms |= {x for x, y in msg.guild.voice_client.channel.permissions_for(msg.author) if y}
            # if (not user_perms >= self.perms or self.bot_maintainers_only) \
            #         and msg.author.id not in s.config_manager.config["global"].get("bot_maintainers", []):
            #     raise UserPermissionError
            if self.perms.check_permissions(msg.author, msg.channel):
                return await f(s, msg)
            else:
                raise UserPermissionError

        wrapped._command = True
        wrapped.aliases = self.aliases
        wrapped.__doc__ = self.doc
        wrapped.name = self.name
        wrapped.perms = self.perms
        wrapped.syntax = self.human_syntax
        wrapped.priority = self.priority
        wrapped.delete_call = self.delete_call
        wrapped.run_anywhere = self.run_anywhere
        wrapped.dm_command = self.dm_command
        wrapped.category = self.category
        return wrapped


class CommandPermissions:
    def __init__(self, permissions_all, permissions_any=None, role_overrides=None, user_overrides=None,
                 optional_permissions=None, bot_maintainers_only=False):
        if permissions_any is None:
            permissions_any = set()
        if role_overrides is None:
            role_overrides = set()
        if user_overrides is None:
            user_overrides = set()
        if optional_permissions is None:
            optional_permissions = {}
        self.permissions_all = permissions_all
        self.permissions_any = permissions_any
        self.role_overrides = role_overrides
        self.user_overrides = user_overrides
        self.optional_permissions = optional_permissions
        self.bot_maintainers = []
        self.bot_maintainers_only = bot_maintainers_only

    @classmethod
    def from_existing(cls, existing_obj: CommandPermissions, permissions_all=None, permissions_any=None,
                      role_overrides=None, user_overrides=None, optional_permissions=None):
        if permissions_all is None:
            permissions_all = existing_obj.permissions_all
        if permissions_any is None:
            permissions_any = existing_obj.permissions_any
        if role_overrides is None:
            role_overrides = existing_obj.role_overrides
        if user_overrides is None:
            user_overrides = existing_obj.user_overrides
        if optional_permissions is None:
            optional_permissions = existing_obj.optional_permissions
        return cls(permissions_all, permissions_any, role_overrides, user_overrides, optional_permissions,
                   bot_maintainers_only=existing_obj.bot_maintainers_only)

    def update(self, new_perms: CommandPermissions):
        self.permissions_all = new_perms.permissions_all
        self.permissions_any = new_perms.permissions_any
        self.role_overrides = new_perms.role_overrides
        self.user_overrides = new_perms.user_overrides
        self.optional_permissions = new_perms.optional_permissions

    def check_permissions(self, member: discord.Member, channel: discord.abc.GuildChannel) -> bool:
        if member.id in self.bot_maintainers:
            return True
        elif self.bot_maintainers_only:
            return False

        if member.id in self.user_overrides:
            return True
        if not {x.id for x in member.roles}.isdisjoint(self.role_overrides):
            return True

        member_permissions_set = {x for x, y in channel.permissions_for(member) if y}
        if member.guild.voice_client:
            member_permissions_set |= {x for x, y in member.guild.voice_client.channel.permissions_for(member) if y}

        if self.permissions_all and not (member_permissions_set > self.permissions_all):
            return False
        if self.permissions_any and member_permissions_set.isdisjoint(self.permissions_any):
            return False
        return True

    def check_optional_permissions(self, optional_permission_set: str, member: discord.Member,
                                   channel: discord.abc.GuildChannel) -> bool:
        if optional_permission_set not in self.optional_permissions:
            raise SyntaxError(f"No such permission set {optional_permission_set}")

        if member.id in self.bot_maintainers:
            return True

        if member.id in self.user_overrides:
            return True
        if not {x.id for x in member.roles}.isdisjoint(self.role_overrides):
            return True

        member_permissions_set = {x for x, y in channel.permissions_for(member) if y}
        if member.guild.voice_client:
            member_permissions_set |= {x for x, y in member.guild.voice_client.channel.permissions_for(member) if y}

        if member_permissions_set > self.optional_permissions[optional_permission_set]:
            return True
        return False



import datetime
import logging
import discord
from argparse import Namespace
from pathlib import Path
from sys import exc_info
from red_star.config_manager import ConfigManager
from red_star.plugin_manager import PluginManager


class RedStar(discord.AutoShardedClient):
    def __init__(self, storage_dir: Path, argv: Namespace):
        self.logger = logging.getLogger("red_star")
        dpy_logger = logging.getLogger("discord")
        if argv.verbose > 0:
            self.logger.setLevel(logging.DEBUG)
            dpy_logger.setLevel(logging.DEBUG if argv.verbose >= 2 else logging.INFO)
        else:
            self.logger.setLevel(logging.INFO)
            dpy_logger.setLevel(logging.INFO)
        self.logger.info("Initializing...")

        allowed_mentions = discord.AllowedMentions(everyone=False, replied_user=False, users=True, roles=True)
        intents = discord.Intents.default()
        intents.members = True
        intents.message_content = True
        super().__init__(intents=intents, allowed_mentions=allowed_mentions,
                         enable_debug_events=(argv.verbose >= 2))

        self.storage_dir = storage_dir
        self.plugin_directories = [Path.cwd() / "plugins"]
        if not argv.portable:
            self.plugin_directories.append(self.storage_dir / "plugins")

        self.config_manager = ConfigManager(storage_dir / "config", storage_dir / "storage")
        self.config = self.config_manager.config

        self.plugin_manager = PluginManager(self)
        self.plugin_manager.load_all_plugins(self.plugin_directories)

        self.logged_in = False
        self.last_error = None

    async def on_ready(self):
        if not self.logged_in:
            self.logged_in = True
            self.logger.info("Logged in as:")
            self.logger.info(self.user)
            self.logger.info(self.user.id)
            self.logger.info("------------")
            self.logger.info("Activating plugins.")
            # create_task(self.global_tick_dispatcher())
            if len(self.guilds) == 0:
                self.logger.info("It looks like you haven't yet added the bot to any servers. Paste the link below "
                                 "into your browser and invite the bot to some servers!")
                self.logger.info(discord.utils.oauth_url(self.user.id))
            await self.plugin_manager.activate_all()

    async def close(self):
        self.logger.warning("Logging out and shutting down.")
        await self.plugin_manager.deactivate_all()
        self.config_manager.save_config()
        await super().close()

    async def on_error(self, event_method: str, *args, **kwargs):
        exc = exc_info()
        self.last_error = exc
        self.logger.exception(f"Unhandled {exc[0].__name__} occurred in {event_method}: ", exc_info=True)

    async def on_typing(self, channel: discord.abc.Messageable, user: discord.abc.User,
                        when: datetime.datetime):
        if not isinstance(channel, discord.abc.GuildChannel):
            return
        await self.plugin_manager.hook_event("on_typing", channel.guild, channel, user, when)

    async def on_message(self, msg: discord.Message):
        if msg.guild is not None:
            await self.plugin_manager.hook_event("on_message", msg.guild, msg)

    async def on_message_delete(self, msg: discord.Message):
        if msg.guild is None:
            return
        await self.plugin_manager.hook_event("on_message_delete", msg.guild, msg)

    async def on_message_edit(self, before: discord.Message, after: discord.Message):
        if after.guild is None:
            return
        await self.plugin_manager.hook_event("on_message_edit", after.guild, before, after)

    async def on_reaction_add(self, reaction: discord.Reaction, user: discord.abc.User):
        if reaction.message.guild is None:
            return
        await self.plugin_manager.hook_event("on_reaction_add", reaction.message.guild, reaction, user)

    async def on_raw_reaction_add(self, payload: discord.RawReactionActionEvent):
        if payload.channel_id is not None:
            channel = self.get_channel(payload.channel_id)
            if isinstance(channel, discord.abc.GuildChannel):
                await self.plugin_manager.hook_event("on_raw_reaction_add", channel.guild, payload)

    async def on_reaction_remove(self, reaction: discord.Reaction, user: discord.abc.User):
        if reaction.message.guild is None:
            return
        await self.plugin_manager.hook_event("on_reaction_remove", reaction, user)

    async def on_raw_reaction_remove(self, payload: discord.RawReactionActionEvent):
        if payload.channel_id is not None:
            channel = self.get_channel(payload.channel_id)
            if isinstance(channel, discord.abc.GuildChannel):
                await self.plugin_manager.hook_event("on_raw_reaction_remove", channel.guild, payload)

    async def on_reaction_clear(self, message: discord.Message, reactions: list[discord.Reaction]):
        if message.guild is None:
            return
        await self.plugin_manager.hook_event("on_reaction_clear", message.guild, message, reactions)

    async def on_guild_channel_create(self, channel: discord.abc.GuildChannel):
        await self.plugin_manager.hook_event("on_guild_channel_create", channel.guild, channel)

    async def on_guild_channel_delete(self, channel: discord.abc.GuildChannel):
        await self.plugin_manager.hook_event("on_guild_channel_delete", channel.guild, channel)

    async def on_guild_channel_update(self, before: discord.abc.GuildChannel, after: discord.abc.GuildChannel):
        await self.plugin_manager.hook_event("on_guild_channel_update", after.guild, before, after)

    async def on_guild_channel_pins_update(self, channel: discord.abc.GuildChannel, last_pin: datetime.datetime):
        await self.plugin_manager.hook_event("on_guild_channel_pins_update", channel.guild, channel, last_pin)

    async def on_member_join(self, member: discord.Member):
        await self.plugin_manager.hook_event("on_member_join", member.guild, member)

    async def on_member_remove(self, member: discord.Member):
        await self.plugin_manager.hook_event("on_member_remove", member.guild, member)

    async def on_member_update(self, before: discord.Member, after: discord.Member):
        await self.plugin_manager.hook_event("on_member_update", after.guild, before, after)

    async def on_guild_join(self, guild: discord.Guild):
        await self.plugin_manager.activate_server_plugins(guild)
        await self.plugin_manager.hook_event("on_guild_join", guild)

    async def on_guild_remove(self, guild: discord.Guild):
        await self.plugin_manager.hook_event("on_guild_remove", guild)
        await self.plugin_manager.deactivate_server_plugins(guild)

    async def on_guild_update(self, before: discord.Guild, after: discord.Guild):
        await self.plugin_manager.hook_event("on_guild_update", after, before, after)

    async def on_guild_role_create(self, role: discord.Role):
        await self.plugin_manager.hook_event("on_guild_role_create", role.guild, role)

    async def on_guild_role_delete(self, role: discord.Role):
        await self.plugin_manager.hook_event("on_guild_role_delete", role.guild, role)

    async def on_guild_role_update(self, before: discord.Role, after: discord.Role):
        await self.plugin_manager.hook_event("on_guild_role_update", after.guild, before, after)

    async def on_guild_emojis_update(self, guild: discord.Guild, before: discord.Emoji, after: discord.Emoji):
        await self.plugin_manager.hook_event("on_guild_emojis_update", guild, before, after)

    async def on_guild_available(self, guild: discord.Guild):
        await self.plugin_manager.hook_event("on_guild_available", guild)

    async def on_guild_unavailable(self, guild: discord.Guild):
        await self.plugin_manager.hook_event("on_guild_unavailable", guild)

    async def on_voice_state_update(self, member: discord.Member, before: discord.VoiceState,
                                    after: discord.VoiceState):
        await self.plugin_manager.hook_event("on_voice_state_update", member.guild, member, before, after)

    async def on_member_ban(self, guild: discord.Guild, member: discord.Member):
        await self.plugin_manager.hook_event("on_member_ban", guild, member)

    async def on_member_unban(self, guild: discord.Guild, member: discord.Member):
        await self.plugin_manager.hook_event("on_member_unban", guild, member)



from __future__ import annotations
import json
import logging
import shutil
import sys
from pathlib import Path
from shutil import copyfile

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Optional
    import discord
    from red_star.plugin_manager import BasePlugin


JsonValues = None | bool | str | int | float | list | dict


class ConfigManager:
    """
    Manages the loading and modification of the configuration files.
    """
    def __init__(self, config_path: Path, storage_path: Path):
        self.logger = logging.getLogger("red_star.config_manager")
        self.logger.debug("Initialized config manager.")
        self.config = {}
        self.config_path = config_path
        self.config_file_path = config_path / "config.json"
        self.storage_path = storage_path
        self.storage_files = {}
        self.load_config()

    def load_config(self):
        temp_path = Path(str(self.config_file_path) + "_bak")
        self.logger.debug("Loading configuration...")
        try:
            with self.config_file_path.open(encoding="utf-8") as fd:
                self.config = json.load(fd)
        except FileNotFoundError:
            if temp_path.exists():
                temp_path.rename(self.config_file_path)
                self.logger.warning(f"Couldn't find {self.config_file_path}!\n"
                                    f"A backup from an interrupted save was found. Attempting to load...")
                return self.load_config()
            self.logger.warning(f"Couldn't find {self.config_file_path}! Copying default configuration...")
            default_path = Path.cwd() / "_default_files/config.json.default"
            self.config_path.mkdir(parents=True, exist_ok=True)
            copyfile(str(default_path), str(self.config_file_path))
            self.logger.info(f"A default configuration has been copied to {self.config_path}.\n"
                             f"Please configure the bot before continuing.")
            sys.exit(1)
        except json.decoder.JSONDecodeError:
            if temp_path.exists():
                self.config_file_path.unlink()
                temp_path.rename(self.config_file_path)
                self.logger.warning(f"The configuration file located at {self.config_file_path} is invalid!\n"
                                    f"A backup from an interrupted save was found. Attempting to load...")
                return self.load_config()
            self.logger.exception(f"The configuration file located at {self.config_file_path} is invalid!\n"
                                  f"Please correct the error below and restart.", exc_info=True)
            sys.exit(1)

        if self.config.get("global", {}).get("__config_version", 0) < 2:
            self._port_config_to_v2()

    def _port_config_to_v2(self):
        # Function to reorganize from plugin-first hierarchy to server-first hierarchy. Plugins that use their own
        # config files will have to manage this themselves.
        self.logger.warning("Porting configuration to newer format. Backup will be created.")
        backup_path = self.config_file_path.with_stem(self.config_file_path.stem + "_old_v1")
        shutil.copyfile(str(self.config_file_path), str(backup_path))
        new_config = {
            "global": {
                "__config_version": 2,
                "bot_maintainers": self.config["bot_maintainers"],
                "token": self.config["token"]
            },
            "default": {}
        }
        for plugin, config in self.config["plugins"].items():
            for k, v in config.items():
                if k.isdigit():
                    new_config.setdefault(k, {})[plugin] = v
                elif k == "default":
                    new_config["default"][plugin] = v
                else:
                    new_config["global"].setdefault(plugin, {})[k] = v
        self.config = new_config
        self.save_config()

    def save_config(self):
        temp_path = Path(str(self.config_file_path) + "_bak")
        with temp_path.open("w", encoding="utf-8") as f:
            json.dump(self.config, f, sort_keys=True, indent=2)
        self.config_file_path.unlink()
        temp_path.rename(self.config_file_path)
        for guild_files in self.storage_files.values():
            for file in guild_files.values():
                file.save()
        self.logger.debug("Saved config files.")

    def get_global_config(self, plugin: str, default_config=None):
        if default_config is None:
            default_config = {}
        default_config = default_config | self.config["global"].setdefault(plugin, {})
        global_config = ConfigDict(self.config["global"].setdefault(plugin, default_config), default_config)
        self.config["global"][plugin] = global_config
        return global_config

    def get_server_config(self, guild: discord.Guild, plugin: str, default_config=None):
        if default_config is None:
            default_config = {}
        default_config |= self.config["default"].setdefault(plugin, default_config)
        server_config = ConfigDict(self.config.setdefault(str(guild.id), {}).setdefault(plugin, default_config),
                                   default_config)
        self.config[str(guild.id)][plugin] = server_config
        self.config["default"][plugin] = default_config
        return server_config

    def get_plugin_storage(self, plugin: BasePlugin) -> PluginStorageFile:
        guild_id = str(plugin.guild.id)
        guild_storage_path = self.storage_path / guild_id
        guild_storage_path.mkdir(parents=True, exist_ok=True)
        filename = guild_storage_path / (plugin.name + ".json")

        storage_file = PluginStorageFile(filename, json_save_args=plugin.storage_save_args(),
                                         json_load_args=plugin.storage_load_args())
        self.storage_files.setdefault(guild_id, {})[plugin.name] = storage_file
        return storage_file

    def save_all_plugin_storage(self):
        for guild in self.storage_files:
            for file in guild:
                file.save()

    def is_maintainer(self, user: discord.abc.User):
        return user.id in self.config.get('bot_maintainers', [])

# Utility classes


class ConfigDict(dict):

    def __init__(self, config: dict[str, JsonValues], default_config: dict[str, JsonValues]):
        super().__init__()
        new_dict = default_config | config
        for k, v in new_dict.items():
            if isinstance(v, dict):
                self[k] = ConfigDict(v, default_config.get(k, {}))
            else:
                self[k] = v


class PluginStorageFile:
    def __init__(self, path: Path, json_save_args: Optional[dict] = None, json_load_args: Optional[dict] = None):
        self.path = path
        self.json_save_args = {} if json_save_args is None else json_save_args
        self.json_load_args = {} if json_load_args is None else json_load_args
        self.contents: JsonValues = {}
        self.load()

    def __repr__(self):
        return f"<PluginStorageFile ({self.path})>"

    def load(self):
        if self.path.exists():
            with self.path.open(encoding="utf-8") as fp:
                self.contents = json.load(fp, **self.json_load_args)

    def save(self):
        if self.contents:
            with self.path.open("w", encoding="utf-8") as fp:
                json.dump(self.contents, fp, **self.json_save_args)
